@@ -37,6 +37,7 @@
 
 #include <libxfce4ui/libxfce4ui.h>
 
+#include <gdk/gdk.h>
 #ifdef GDK_WINDOWING_X11
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
@@ -55,13 +56,6 @@
 
 
 static void     terminal_app_finalize                 (GObject            *object);
-static void     terminal_app_update_accels            (TerminalApp        *app);
-static void     terminal_app_update_tab_key_accels    (TerminalApp        *app);
-static void     terminal_app_store_tab_key_accel      (gpointer            data,
-                                                       const gchar        *accel_path,
-                                                       guint               accel_key,
-                                                       GdkModifierType     accel_mods,
-                                                       gboolean            changed);
 static void     terminal_app_update_windows_accels    (gpointer            user_data);
 static gboolean terminal_app_accel_map_load           (gpointer            user_data);
 static gboolean terminal_app_accel_map_save           (gpointer            user_data);
@@ -101,7 +95,6 @@ struct _TerminalApp
   guint                accel_map_load_id;
   guint                accel_map_save_id;
   GtkAccelMap         *accel_map;
-  GSList              *tab_key_accels;
 };
 
 
@@ -136,18 +129,12 @@ static void
 terminal_app_init (TerminalApp *app)
 {
   app->preferences = terminal_preferences_get ();
-  g_signal_connect_swapped (G_OBJECT (app->preferences), "notify::shortcuts-no-menukey",
-                            G_CALLBACK (terminal_app_update_accels), app);
-  g_signal_connect_swapped (G_OBJECT (app->preferences), "notify::shortcuts-no-helpkey",
-                            G_CALLBACK (terminal_app_update_accels), app);
 
   /* remember the original menu bar accel */
   g_object_get (G_OBJECT (gtk_settings_get_default ()),
                 "gtk-menu-bar-accel", &app->initial_menu_bar_accel,
                 NULL);
 
-  terminal_app_update_accels (app);
-
   /* schedule accel map load and update windows when finished */
   app->accel_map_load_id = gdk_threads_add_idle_full (G_PRIORITY_LOW, terminal_app_accel_map_load, app,
                                                       terminal_app_update_windows_accels);
@@ -182,7 +169,6 @@ terminal_app_finalize (GObject *object)
     }
   g_slist_free (app->windows);
 
-  g_signal_handlers_disconnect_by_func (G_OBJECT (app->preferences), G_CALLBACK (terminal_app_update_accels), app);
   g_object_unref (G_OBJECT (app->preferences));
 
   if (app->initial_menu_bar_accel != NULL)
@@ -191,79 +177,11 @@ terminal_app_finalize (GObject *object)
   if (app->session_client != NULL)
     g_object_unref (G_OBJECT (app->session_client));
 
-  for (lp = app->tab_key_accels; lp != NULL; lp = lp->next)
-    g_free (((TerminalAccel*) lp->data)->path);
-  g_slist_free_full (app->tab_key_accels, g_free);
-
   (*G_OBJECT_CLASS (terminal_app_parent_class)->finalize) (object);
 }
 
 
 
-static void
-terminal_app_update_accels (TerminalApp *app)
-{
-  gboolean        no_key;
-  GdkModifierType mod;
-  guint           key;
-
-  g_object_get (G_OBJECT (app->preferences),
-                "shortcuts-no-menukey", &no_key,
-                NULL);
-  g_object_set (G_OBJECT (gtk_settings_get_default ()),
-                "gtk-menu-bar-accel",
-                no_key ? NULL : app->initial_menu_bar_accel,
-                NULL);
-  gtk_accelerator_parse (app->initial_menu_bar_accel, &key, &mod);
-  gtk_accel_map_change_entry ("<Actions>/terminal-window/toggle-menubar",
-                              no_key ? 0 : key, no_key ? 0 : mod, TRUE);
-
-  g_object_get (G_OBJECT (app->preferences),
-                "shortcuts-no-helpkey", &no_key,
-                NULL);
-  gtk_accel_map_change_entry ("<Actions>/terminal-window/contents",
-                              no_key ? 0 : GDK_KEY_F1, 0, TRUE);
-}
-
-
-
-static void
-terminal_app_update_tab_key_accels (TerminalApp *app)
-{
-  if (app->tab_key_accels != NULL)
-    {
-      GSList *lp;
-      for (lp = app->tab_key_accels; lp != NULL; lp = lp->next)
-        g_free (((TerminalAccel*) lp->data)->path);
-      g_slist_free_full (app->tab_key_accels, g_free);
-      app->tab_key_accels = NULL;
-    }
-  gtk_accel_map_foreach (app, terminal_app_store_tab_key_accel);
-}
-
-
-
-static void
-terminal_app_store_tab_key_accel   (gpointer         data,
-                                    const gchar     *accel_path,
-                                    guint            accel_key,
-                                    GdkModifierType  accel_mods,
-                                    gboolean         changed)
-{
-  if (accel_key == GDK_KEY_Tab || accel_key == GDK_KEY_ISO_Left_Tab)
-    {
-      TerminalApp *app = TERMINAL_APP (data);
-      TerminalAccel *accel = g_new0 (TerminalAccel, 1);
-
-      accel->path = g_strdup (g_strrstr (accel_path, "/") + 1); // <Actions>/terminal-window/action
-      accel->mods = accel_mods;
-
-      app->tab_key_accels = g_slist_prepend (app->tab_key_accels, accel);
-    }
-}
-
-
-
 static void
 terminal_app_update_windows_accels (gpointer user_data)
 {
@@ -272,8 +190,6 @@ terminal_app_update_windows_accels (gpointer user_data)
 
   for (lp = app->windows; lp != NULL; lp = lp->next)
     {
-      terminal_window_update_tab_key_accels (TERMINAL_WINDOW (lp->data), app->tab_key_accels);
-
       /* the accel_map is loaded after the first windows are created, so they can't create the go-to actions on page-insert */
       terminal_window_update_goto_accels (TERMINAL_WINDOW (lp->data));
     }
@@ -318,9 +234,6 @@ terminal_app_accel_map_changed (TerminalApp *app)
   /* schedule new save */
   app->accel_map_save_id = gdk_threads_add_timeout_seconds (10, terminal_app_accel_map_save, app);
 
-  /* identify accelerators containing the Tab key */
-  terminal_app_update_tab_key_accels (app);
-
   /* update the tab-key accel list in each window */
   terminal_app_update_windows_accels (app);
 }
@@ -346,12 +259,21 @@ terminal_app_accel_map_load (gpointer user_data)
   g_signal_connect_swapped (G_OBJECT (app->accel_map), "changed",
       G_CALLBACK (terminal_app_accel_map_changed), app);
 
-  /* identify accelerators containing the Tab key */
-  terminal_app_update_tab_key_accels (app);
-
   return FALSE;
 }
 
+void
+terminal_app_load_accels (TerminalApp *app)
+{
+  if (G_UNLIKELY (app->accel_map_load_id != 0))
+    {
+      g_source_remove (app->accel_map_load_id);
+      app->accel_map_load_id = 0;
+    }
+
+  terminal_app_accel_map_load (app);
+}
+
 
 
 static gboolean
@@ -390,8 +312,6 @@ terminal_app_take_window (TerminalApp *app,
   g_signal_connect (G_OBJECT (window), "key-release-event",
                     G_CALLBACK (terminal_app_unset_urgent_bell), app);
   app->windows = g_slist_prepend (app->windows, window);
-
-  terminal_window_update_tab_key_accels (TERMINAL_WINDOW (window), app->tab_key_accels);
 }
 
 
@@ -928,7 +848,7 @@ terminal_app_open_window (TerminalApp        *app,
 #else
       if (!gtk_window_parse_geometry (GTK_WINDOW (window), geometry))
 #endif
-        g_printerr (_("Invalid geometry string \"%s\"\n"), geometry);
+        g_warning ("Invalid geometry string \"%s\"", geometry);
 
       /* cleanup */
       g_free (geometry);
@@ -1008,7 +928,7 @@ terminal_app_process (TerminalApp  *app,
         }
       else
         {
-          g_printerr (_("Failed to connect to session manager: %s\n"), err->message);
+          g_warning ("Failed to connect to session manager: %s", err->message);
           g_error_free (err);
         }
     }
