@@ -389,6 +389,13 @@ terminal_window_dropdown_set_property (GObject      *object,
       return;
 
     case PROP_DROPDOWN_STATUS_ICON:
+      /*
+       * FIXME: GtkStatusIcon is X11 only
+       * See also https://gitlab.xfce.org/apps/xfce4-terminal/-/issues/197
+       */
+      if (!GDK_IS_X11_DISPLAY (gdk_display_get_default ()))
+        return;
+
       if (g_value_get_boolean (value))
         {
           /* icon_name will be NULL initially after creation; will update it later */
@@ -550,9 +557,9 @@ terminal_window_dropdown_status_icon_popup_menu (GtkStatusIcon          *status_
       G_CALLBACK (gtk_widget_destroy), NULL);
 
   xfce_gtk_menu_item_new_from_action_entry (terminal_window_get_action_entry (window, TERMINAL_WINDOW_ACTION_PREFERENCES), G_OBJECT (window), GTK_MENU_SHELL (menu));
-  xfce_gtk_menu_append_seperator (GTK_MENU_SHELL (menu));
+  xfce_gtk_menu_append_separator (GTK_MENU_SHELL (menu));
   xfce_gtk_menu_item_new_from_action_entry (terminal_window_get_action_entry (window, TERMINAL_WINDOW_ACTION_CLOSE_WINDOW), G_OBJECT (window), GTK_MENU_SHELL (menu));
-  xfce_gtk_menu_append_seperator (GTK_MENU_SHELL (menu));
+  xfce_gtk_menu_append_separator (GTK_MENU_SHELL (menu));
 
   gtk_widget_show_all (menu);
   gtk_menu_popup_at_pointer (GTK_MENU (menu), NULL);
